/*---------------------------------------
#          (A)OMUS                      #
#      by Cedric Piette                 #
#           v0.72                       #
#      based on WalkSat                 #
----------------------------------------*/

#define VERSION "AOMUS_v0.72"

/************************************/
/* Compilation flags                */
/************************************/

/* If the constant DYNAMIC is set, then dynamic arrays are used instead of static arrays.
   This allows very large or very small problems to be handled, without
   having to adjust the constants MAXATOM and/or MAXCLAUSE.  However, on some
   architectures the compiled program is about 25% slower, because not
   as many optimizations can be performed. */

#define DYNAMIC 1 
#define BIGINT long long int


/************************************/
/* Standard includes                */
/************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <limits.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>

#include <sys/times.h>
#include <sys/time.h>
#include <unistd.h>


#define random() rand()
#define srandom(seed) srand(seed)


/* Sometimes CLK_TCK is defined in a weird place */
#ifndef CLK_TCK
#define CLK_TCK 60
#endif


/************************************/
/* Constant parameters              */
/************************************/

#define MAXATOM 150000		/* maximum possible number of atoms */


#define STOREBLOCK 2000000	/* size of block to malloc each time */
#define MAXCLAUSE 500000	/* maximum possible number of clauses */


#define TRUE 1
#define FALSE 0

#define MAXLENGTH 500           /* maximum number of literals which can be in any clause */

/************************************/
/* Internal constants               */
/************************************/

enum heuristics { RANDOM, BEST, TABU, NOVELTY, RNOVELTY, NOVELTY_PLUS, 
		  RNOVELTY_PLUS};

#define NOVALUE -1
#define INIT_PARTIAL 1
#define HISTMAX 101		/* length of histogram of tail */

#define Var(CLAUSE, POSITION) (ABS(clause[CLAUSE][POSITION]))

static int scratch;
#define ABS(x) ((scratch=(x))>0?(scratch):(-scratch))

#define BIG 100000000
#define OPTI 111111


/************************************/
/* Main data structures             */
/************************************/

/* Atoms start at 1 */
/* Not a is recorded as -1 * a */
/* One dimensional arrays are statically allocated. */
/* Two dimensional arrays are dynamically allocated in */
/* the second dimension only.  */

int numatom;
int numclause;
int numliterals;

#ifdef DYNAMIC
int ** clause;                   /* clauses to be satisfied */
/* indexed as clause[clause_num][literal_num] */

int * size;			/* length of each clause */
int * false;			/* clauses which are false */
int * lowfalse;
int * wherefalse;		/* where each clause is listed in false */
int * numtruelit;		/* number of true literals in each clause */
#else
int * clause[MAXCLAUSE];	/* clauses to be satisfied */
                 		/* indexed as clause[clause_num][literal_num] */

int size[MAXCLAUSE];		/* length of each clause */
int false[MAXCLAUSE];		/* clauses which are false */
int lowfalse[MAXCLAUSE];
int wherefalse[MAXCLAUSE];	/* where each clause is listed in false */
int numtruelit[MAXCLAUSE];	/* number of true literals in each clause */
#endif

int *occurence[2*MAXATOM+1];	/* where each literal occurs */
				/* indexed as occurence[literal+MAXATOM][occurence_num] */

int numoccurence[2*MAXATOM+1];	/* number of times each literal occurs */


int atom[MAXATOM+1];		/* value of each atom */ 
int lowatom[MAXATOM+1];
int solution[MAXATOM+1];

int nbatom[MAXATOM+1];


int changed[MAXATOM+1];		/* step at which atom was last flipped */

int breakcount[MAXATOM+1];	/* number of clauses that become unsat if var if flipped */
int makecount[MAXATOM+1];	/* number of clauses that become sat if var if flipped */

int numfalse;			/* number of false clauses */

/* MUS search */



#ifdef DYNAMIC

int * clausedirectory;             /* directory used to convert the current search clause's index into */
                                   /* real clause's index */
int * backup;                      /* clause directory used to save previous clausedirectory */
int * falseorcriticmeter;          /* number of false or critical clause */
int ** clauseinocc;                /* where each clause appears in occurence*/
int * defended;                    /* give the defended clauses*/
int * defbackup;                   /* give the defended clauses of the previous loop */
int * remclause;                   /* give the cleared clauses during sic or asmus search  */
int * minfalseasmus;               /* give the minimum number of false clauses when the ième clause is false */

#else

int clausedirectory[MAXCLAUSE];    /* directory used to convert the current search clause's index into */
                                   /* real clause's index */
int backup[MAXCLAUSE];             /* clause directory used to save previous clausedirectory */
int falseorcriticmeter[MAXCLAUSE]; /* number of false or critical clause */
int * clauseinocc[MAXCLAUSE];      /* where each clause appears in occurence*/
int defended[MAXCLAUSE];           /* give the defended clauses*/
int defbackup[MAXCLAUSE];          /* give the defended clauses of the previous loop */
int remclause[MAXCLAUSE];          /* give the cleared clauses during sic search  */
int minfalseasmus[MAXCLAUSE];      /* give the minimum number of false clauses when the ième clause is false */

#endif

int testnumber = 0;                     /* number of test during the try in progress */
int numclauseinit;                      /* initial number of clause */
int criticalpoint;                      /* critical point allowing to know the clauses to be removed  */
int * currentoccurence[2*MAXATOM+1];    /* give the boxes of occurence used for research in progress */
int * poslitinclause[2*MAXATOM+1];      /* give the position of litteral in the clause */
                                        /* this table is complementary to table occurence */
int numoccurenceinitial[2*MAXATOM+1];   /* initial number of litteral occurence */


int numbackup=1;                        /* number of clause in backup table */
int numdefended;                        /* number of clause defended */
int numdefbackup;                       /* number of clause defended into the backup save */
int iteration = 1;                      /* number of try */
int nofound = 1;                        /* allows to know if research were profitable  */
int numvarbackup;                       /* number of variable in the backup save */
int numactuallyvar;                     /* number of actually variable */ 
int firstloop = 1;                     
int finished;                          
int siccpt = 0;                        /* number of MUS found */ 
unsigned int newth;
int withdrawn;
unsigned int raising = MAXCLAUSE;
int optimization;
int nowithdrawnclause;
unsigned int numwithdrawn = 1;
unsigned int level;


int falseClause = FALSE;
int finetune = FALSE;                   
int print = FALSE;
int asmus = FALSE;
int sic = FALSE;



char * backupfile = "backup";

/************************************/
/* Complete method                  */
/************************************/

#define BLOC 4000


int ** stack;
int * sizestack;
int top;
int sizespace;




/************************************/
/* Global flags and parameters      */
/************************************/

int status_flag = 0;		/* value returned from main procedure */
int abort_flag;

int heuristic = BEST;		/* heuristic to be used */
int numerator = NOVALUE;	/* make random flip with numerator/denominator frequency */
int denominator = 100;		
int tabu_length;		/* length of tabu list */

int wp_numerator = NOVALUE;	/* walk probability numerator/denominator */
int wp_denominator = 100;		

BIGINT numflip;		/* number of changes so far */
BIGINT numnullflip;		/*  number of times a clause was picked, but no  */
				/*  variable from it was flipped  */
int numrun = 10;
BIGINT cutoff = 100000;
BIGINT base_cutoff = 100000;
int target = 0;
int numtry = 0;			/* total attempts at solutions */

int makeflag = FALSE;		/* set to true by heuristics that require the make values to be calculated */


/* Initialization options */

char initfile[100] = { 0 };
int initoptions = FALSE;

/* Randomization */

unsigned int seed;  /* Sometimes defined as an unsigned long int */
unsigned int seed2;  /* Sometimes defined as an unsigned long int */


struct timeval tv;
struct timezone tzp;





/************************************/
/* Forward declarations             */
/************************************/

void parse_parameters(int argc,char *argv[]);


int pickrandom(void);
int pickbest(void);
int picktabu(void);
int picknovelty(void);
int pickrnovelty(void);
int picknoveltyplus(void);
int pickrnoveltyplus(void);

char * heuristic_names[] = { "random", "best", "tabu", "novelty", "rnovelty",
			     "novelty+", "rnovelty+"};

static int (*pickcode[])(void) =
{pickrandom, pickbest, picktabu, picknovelty, pickrnovelty,
 picknoveltyplus, pickrnoveltyplus};




void scanone(int argc, char *argv[], int i, int *varptr);
void scanoneu(int argc, char *argv[], int i, unsigned int *varptr);
void scanonell(int argc, char *argv[], int i, BIGINT *varptr);
void init(char initfile[], int initoptions);

void initprob(void); 
void resetprob(void);
void flipatom(int toflip);
void level_init(void);

void handle_interrupt(int sig);

long super(int i);


void  print_Var_Clause(void);
void  init_random_seed(void);
void print_stat_of_try(void);
void print_result(void);

void init_numtry_and_numrun(void);
void initMUSsearch(void);
void initMUSsearch2(void);

void flipatom_and_count_defended(void);
void flipatom_and_count_defended_and_count_minfalseasmus(void);
void (*seek)(void) = flipatom_and_count_defended;

void evaluate_cutoff(void);

void inc_false_meter(void);
void inc_critical_meter(void);
void (*incfalseorcriticmeter)(void) = inc_critical_meter;

void removeclause(void);
void remove_parasite(int * parasiticatoms , int nbpar);
void removeclause_and_parasite(void);
void (*rmclause)(void) = removeclause;

void nullmethod(void);
void agingmethod(void);
void (*resetmeter)(void) = nullmethod;

void maxminmeter2(int * max,int * min);
void maxminmeter1(int * max,int * min);
void oldthresholdcalculator(void);
void newthresholdcalculator(void);
void (*thresholdcalculator) (void) = oldthresholdcalculator;

void resetclause(void);
int isCritical(int clause);
char * entierEnChaine(int n);
void directory_into_file(char * file,const int * directory,const int taille);
void DL_verification(void);
void fineTune_verification(int * directory,int * num,int index);
void control_of_end_of_try_loop(void);
void resultat(int ret);
void fineTune_extraction(void);
void print_try(void);
void allot_name_file(char ** file);
void reset_for_SIC_search(void);
void reset_for_ASMUS_search(void);
int unsat(char * file);
int clause_of_the_lowest_layer(void);
void back_return(void);

void correction_clause(void);
void init_stack(void);
void push(int * cl, int sz);
void pop(int ** cl, int * sz);
void quicksort(int sz,int * dirsort);

void reset_sic(void);
void reset_asmus(void);
void (*reset)(void) = resetprob;



/************************************/
/* Main                             */
/************************************/


int main(int argc,char *argv[])
{
  srand(time(NULL));
  seed = rand();
  level = 4;
  parse_parameters(argc, argv);
  printf("seed = %d\n",seed);
  level_init();
  initprob();
  signal(SIGINT, handle_interrupt);
  abort_flag = FALSE;
  nowithdrawnclause = 0;
  init_stack();
  do{
     print_Var_Clause();  
     finished = 0;
     optimization = FALSE;
     print_try();
     while (! abort_flag && !finished) {
        resetmeter();
        init_numtry_and_numrun();
        seed2=seed;
        numfalse = target + 1;
        evaluate_cutoff();
        while (numfalse > target && numtry < numrun ) {    
           init(initfile, initoptions);	
           numtry++;      
           numflip = 0;
           while((numfalse > target) && (numflip < cutoff)){
              numflip++;
              seek();
           }
        }  
        control_of_end_of_try_loop();
        print_stat_of_try();   
        if(finished==0 || finished == 2) resetclause();
        if(finished==3){
           back_return();
           finished = 0;
        }
     }
     DL_verification();
     reset();
  }while(nofound);
  print_result();
  return status_flag;
}



void parse_parameters(int argc,char *argv[])
{
  int i;
  int temp;

  for (i=1;i < argc;i++)
    {
      if (strcmp(argv[i],"-seed") == 0)
	if (i+1>=argc || sscanf(argv[i+1],"%u",&seed)!=1){
	  init_random_seed(); 
	}else{
	  ++i;
	}
      else if (strcmp(argv[i],"-status") == 0)
	status_flag = 1;
      else if (strcmp(argv[i],"-random") == 0)
	heuristic = RANDOM;
      else if (strcmp(argv[i],"-novelty") == 0){
	heuristic = NOVELTY;
	makeflag = TRUE;
      }
      else if (strcmp(argv[i],"-rnovelty") == 0){
	heuristic = RNOVELTY;
	makeflag = TRUE;
      }
      else if (strcmp(argv[i],"-novelty+") == 0){
	heuristic = NOVELTY_PLUS;
	makeflag = TRUE;
      }
      else if (strcmp(argv[i],"-rnovelty+") == 0){
	heuristic = RNOVELTY_PLUS;
	makeflag = TRUE;
      }
      else if (strcmp(argv[i],"-best") == 0)
	heuristic = BEST;
      else if (strcmp(argv[i],"-noise") == 0){
	scanone(argc,argv,++i,&numerator);
	if (i < argc-1 && sscanf(argv[i+1],"%i",&temp)==1){
	  denominator = temp;
	  i++;
	}
      }
      else if (strcmp(argv[i],"-wp") == 0){
	scanone(argc,argv,++i,&wp_numerator);
	if (i < argc-1 && sscanf(argv[i+1],"%i",&temp)==1){
	  wp_denominator = temp;
	  i++;
	}
      }
      else if (strcmp(argv[i],"-init") == 0  && i < argc-1)
	sscanf(argv[++i], " %s", initfile);
      else if (strcmp(argv[i],"-partial") == 0)
	initoptions = INIT_PARTIAL;
      else if (strcmp(argv[i],"-tabu") == 0)
	{
	  scanone(argc,argv,++i,&tabu_length);
	  heuristic = TABU;
	}
      else if(strcmp(argv[i],"-false") == 0){
	incfalseorcriticmeter = inc_false_meter;
	falseClause = TRUE;
      }
      else if(strcmp(argv[i],"-fineTune") == 0)
	finetune = TRUE;
      else if(strcmp(argv[i],"-print") == 0)
	print = TRUE;
      else if(strcmp(argv[i],"-parasite") == 0)
	rmclause = removeclause_and_parasite;
      else if(strcmp(argv[i],"-raising") == 0){
	scanoneu(argc,argv,++i,&raising);
	thresholdcalculator = newthresholdcalculator;
      }
      else if(strcmp(argv[i],"-aging") ==0)
	resetmeter = agingmethod;
      else if(strcmp(argv[i],"-newth") == 0){
	thresholdcalculator = newthresholdcalculator;
	scanoneu(argc,argv,++i,&newth);
      }
      else if(strcmp(argv[i],"-withdrawn") == 0){
	scanoneu(argc,argv,++i,&numwithdrawn);
      }
      else if(strcmp(argv[i],"-level") == 0){
	scanoneu(argc,argv,++i,&level);
      }
      else if(strcmp(argv[i],"-SIC") == 0 && asmus != TRUE){
	sic = TRUE;
	reset = reset_sic;
	printf("SIC OPTION ACTIVATED\n");
      }
      else if(strcmp(argv[i],"-ASMUS") == 0 && sic != TRUE){
	asmus = TRUE;
	seek = flipatom_and_count_defended_and_count_minfalseasmus;
	reset = reset_asmus;
	printf("ASMUS OPTION ACTIVATED\n");
      }
      else if( (strcmp(argv[i],"-file") == 0) && ( argv[i+1] !=NULL))
	backupfile = argv[++i];
  else 
	{
           fprintf(stderr,"----- (A)OMUS  v0.72 -----\n\n");
           fprintf(stderr, "General parameters:\n");
           fprintf(stderr, "  -seed N = fix the seed to 'N'. \n");
           fprintf(stderr, "            default : the seed is initialised randomly\n");
           fprintf(stderr, "  -print = print different informations during the computation\n");
           fprintf(stderr, "  -fineTune = enable to extract an exact MUS instead of an approximation (OMUS procedure)\n");
           fprintf(stderr, "  -file FILE = save the obtained subformula in 'FILE' (default 'backup')\n\n\n");
           fprintf(stderr, "Experimental parameters:\n");
           fprintf(stderr, "  -parasite = remove parasitic clauses\n");
           fprintf(stderr, "  -aging = uses an aging algorithm to count critical clauses \n");
           fprintf(stderr, "  -newth N = uses an other method to choose critical threshold with N coefficient\n");
           fprintf(stderr, "  -raising N = remove N clauses (maximum) at each step\n\n\n");
           fprintf(stderr, "Heuristics:\n");
           fprintf(stderr, "  -random -best -tabu N -novelty -rnovelty\n");
           fprintf(stderr, "  -noise N or -noise N M (default M = 100)\n");
           fprintf(stderr, "  -novelty+ -rnovelty+\n");
           fprintf(stderr, "  -wp N or -wp N M (default M = 100) = cycle breaking for (r)novelty+\n\n");
           exit(-1);
	}     
    }
  base_cutoff = cutoff;

  if (numerator==NOVALUE){
    switch(heuristic) {
    case BEST:
    case NOVELTY:
    case RNOVELTY:
    case NOVELTY_PLUS:
    case RNOVELTY_PLUS:
      numerator = 50;
      break;
    default:
      numerator = 0;
      break;
    }
  }
  if (wp_numerator==NOVALUE){
    switch(heuristic) {
    case NOVELTY_PLUS:
    case RNOVELTY_PLUS:
      wp_numerator = 1;
      break;
    default:
      wp_numerator = 0;
      break;
    }
  }
}



long super(int i)
{
  long power;
  int k;

  if (i<=0){
    fprintf(stderr, "bad argument super(%d)\n", i);
    exit(1);
  }
  /* let 2^k be the least power of 2 >= (i+1) */
  k = 1;
  power = 2;
  while (power < (i+1)){
    k += 1;
    power *= 2;
  }
  if (power == (i+1)) return (power/2);
  return (super(i - (power/2) + 1));
}

void handle_interrupt(int sig)
{
  if (abort_flag) exit(-1);
  abort_flag = TRUE;
}

void scanone(int argc, char *argv[], int i, int *varptr)
{
  if (i>=argc || sscanf(argv[i],"%i",varptr)!=1){
    fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
    exit(-1);
  }
}

void scanoneu(int argc, char *argv[], int i, unsigned int *varptr)
{
  if (i>=argc || sscanf(argv[i],"%u",varptr)!=1){
    fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
    exit(-1);
  }
}

void scanonell(int argc, char *argv[], int i, BIGINT *varptr)
{
  if (i>=argc || sscanf(argv[i],"%Li",varptr)!=1){
    fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
    exit(-1);
  }
}


void init(char initfile[], int initoptions)
{
  int i;
  int j;
  int k;
  int thetruelit;
  FILE * infile;
  int lit;
  for(i = 0, k = clausedirectory[i] ;i < numclause;k = clausedirectory[++i]){
    numtruelit[k] = 0;
  }
  numfalse = 0;
  for(i = 1;i < numatom+1;i++)
    {
      changed[i] = -BIG;
      breakcount[i] = 0;
      makecount[i] = 0;
      nbatom[i]=0;
    }
  srandom(seed2++);
  if (initfile[0] && initoptions!=INIT_PARTIAL){
    for(i = 1;i < numatom+1;i++)
      atom[i] = 0;
  }
  else {
    for(i = 1;i < numatom+1;i++)
      atom[i] = random()%2;
  }
  if (initfile[0]){
    if ((infile = fopen(initfile, "r")) == NULL){
      fprintf(stderr, "Cannot open %s\n", initfile);
      exit(1);
    }
    i=0;
    while (fscanf(infile, " %d", &lit)==1){
      i++;
      if (ABS(lit)>numatom){
	fprintf(stderr, "Bad init file %s\n", initfile);
	exit(1);
      }
      if (lit<0) atom[-lit]=0;
      else atom[lit]=1;
    }
    if (i==0){
      fprintf(stderr, "Bad init file %s\n", initfile);
      exit(1);
    }
    fclose(infile);
  }
  /* Initialize breakcount and makecount in the following: */
  for(k = 0, i = clausedirectory[k];k < numclause; i = clausedirectory[++k] )
    {
      for(j = 0;j < size[i];j++)
	{
	
	  if((clause[i][j]> 0) == atom[ABS(clause[i][j])])
	    {
	      numtruelit[i]++;
	      thetruelit = clause[i][j];
	    }
	}
   
      if(numtruelit[i] == 0)
	{
	  wherefalse[i] = numfalse;
	  false[numfalse] = i;
	  numfalse++;
	  for(j = 0;j < size[i];j++){
	    makecount[ABS(clause[i][j])]++;
	  }
	}
      else if (numtruelit[i] == 1)
	{
	  breakcount[ABS(thetruelit)]++;
	}
    }
}


void resetprob(void)
{
 
  int i;
  int j;
  int numocc;


  for(i = 1;i <= numatom;i++){
    numoccurence[MAXATOM + i] = 0;
    numoccurence[MAXATOM - i] = 0;
  }

  numdefended = 0;
  numclause = numclauseinit;
  numactuallyvar = numatom;

  for(i = 0;i < numclause;i++)
    {

      for(j = 0;j < size[i];j++)
	{
	  clausedirectory[i] = i;
	  defended[i] = 0;
	  occurence[clause[i][j]+MAXATOM]
	    [numoccurence[clause[i][j]+MAXATOM]] = i;
	      
	  clauseinocc[i][j] = numoccurence[clause[i][j]+MAXATOM]  ;
	     
	  poslitinclause[clause[i][j]+MAXATOM]
	    [numoccurence[clause[i][j]+MAXATOM]]= j;
	
	  ++numoccurence[clause[i][j]+MAXATOM];
	    	   
	}
	 
    }

  for(i = 1;i <= numatom ;++i)
    {
      numocc = numoccurence[MAXATOM + i];
      for(j = 0;j <= numocc;++j)
	currentoccurence[MAXATOM + i][j] = j;
      numocc = numoccurence[MAXATOM - i];
      for(j = 0;j <= numocc;++j)
	currentoccurence[MAXATOM - i][j] = j;

    }

 
 


}

void evaluate_cutoff(){
   cutoff = (long int) ((numactuallyvar<<2)*sqrt((float) (numclause<<2))*super(iteration));
   if (cutoff<5000) {cutoff=cutoff<<1;}
}


void initprob(void)
{
  
  int i;
  int j;
  int k;
  int lastc;
  int nextc;
  int *storeptr;
  int freestore;
  int lit;

  while ((lastc = getchar()) == 'c')
    {
      while ((nextc = getchar()) != EOF && nextc != '\n');
    }
  ungetc(lastc,stdin);
  if (scanf("p cnf %i %i",&numatom,&numclause) != 2)
    {
      fprintf(stderr,"Bad input file\n");
      exit(-1);
    }
  if(numatom > MAXATOM)
    {
      fprintf(stderr,"ERROR - too many atoms\n");
      exit(-1);
    }



#ifdef DYNAMIC
  clause = (int **) malloc(sizeof(int *)*(numclause+1));
  size = (int *) malloc(sizeof(int)*(numclause+1));
  false = (int *) malloc(sizeof(int)*(numclause+1));
  lowfalse = (int *) malloc(sizeof(int)*(numclause+1));
  wherefalse = (int *) malloc(sizeof(int)*(numclause+1));
  numtruelit = (int *) malloc(sizeof(int)*(numclause+1));
#else
  if(numclause > MAXCLAUSE)                     
    {                                      
      fprintf(stderr,"ERROR - too many clauses\n"); 
      exit(-1);                              
    }                                        
#endif

 
  freestore = 0;
  numliterals = 0;
  for(i = 0;i < 2*MAXATOM+1;i++)
    numoccurence[i] = 0;
  for(i = 0;i < numclause;i++)
    {
      size[i] = -1;
      if (freestore < MAXLENGTH)
	{
	  storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
	  freestore = STOREBLOCK;
	  //	  fprintf(stderr,"allocating memory...\n");
	}

      clause[i]= storeptr;

      do
	{
	  size[i]++;
	  if(size[i] > MAXLENGTH)
	    {
	      printf("ERROR - clause too long\n");
	      exit(-1);
	    }
	  if (scanf("%i ",&lit) != 1)
	    {
	      fprintf(stderr, "Bad input file\n");
	      exit(-1);
	    }
	  if(lit != 0)
	    {
	      *(storeptr++) = lit;
	      freestore--;
	      numliterals++;
	      numoccurence[lit+MAXATOM]++;
	    }
	}
      while(lit != 0);
    }


  if(size[0] == 0)
    {
      fprintf(stderr,"ERROR - incorrect problem format or extraneous characters\n");
      exit(-1);
    }

  initMUSsearch();
  initMUSsearch2();


  for(i = 0;i < 2*MAXATOM+1;i++)
    {
      if (freestore < numoccurence[i])
	{
	  storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
	  freestore = STOREBLOCK;

	}
      occurence[i] = storeptr;
      freestore -= numoccurence[i];
      storeptr += numoccurence[i];
      numoccurence[i] = 0;
    }
  
  

  for(i = 0;i < numclause;i++)
    {
      falseorcriticmeter[i] = 0;
      for(j = 0;j < size[i];j++)
	{
	  occurence[clause[i][j]+MAXATOM]
	    [numoccurence[clause[i][j]+MAXATOM]] = i;
	      
	  clauseinocc[i][j] = numoccurence[clause[i][j]+MAXATOM]  ;
	     
	 
	  poslitinclause[clause[i][j]+MAXATOM]
	    [numoccurence[clause[i][j]+MAXATOM]]= j;
	    
	  ++numoccurenceinitial[clause[i][j]+MAXATOM];

	  ++numoccurence[clause[i][j]+MAXATOM];
	    	   
	}

	 
    }

 
  if(asmus == TRUE)
    {
      for(k = 0, i = clausedirectory[k];k < numclause; i = clausedirectory[++k] )
	minfalseasmus[i] = numclause;
     
    }

 

  



}


/*
  the "standard" flipatom implementation
*/
void flipatom(int toflip)
{
  int i, j, k;			
  int toenforce;		
  register int cli;
  register int lit;
  int numocc;
  register int sz;
  register int * litptr;
  int * occptr;
  int * occptrind;


  if (toflip == NOVALUE){
    numnullflip++;
    return;
  }

  changed[toflip] = numflip;
  if(atom[toflip] > 0)
    toenforce = -toflip;
  else
    toenforce = toflip;
  atom[toflip] = 1-atom[toflip];

  numocc = numoccurence[MAXATOM-toenforce];
  occptr = occurence[MAXATOM-toenforce];

  occptrind = currentoccurence[MAXATOM-toenforce];


  for(k = 0,i = occptrind[0]; k < numocc ; ++k, i=occptrind[i+1] )
    {

      
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = occptr[i];

  

      if (--numtruelit[cli] == 0){
	false[numfalse] = cli;
	wherefalse[cli] = numfalse;
	numfalse++;
	/* Decrement toflip's breakcount */
	breakcount[toflip]--;

	if (makeflag){
	  /* Increment the makecount of all vars in the clause */
	  sz = size[cli];
	  litptr = clause[cli];
	  for (j=0; j<sz; j++){
	    lit = clause[cli][j]; 
	    makecount[ABS(lit)]++;
	  }
	}
      }
      else if (numtruelit[cli] == 1){
	/* Find the lit in this clause that makes it true, and inc its breakcount */
	sz = size[cli];
	litptr = clause[cli];
	for (j=0; j<sz; j++){
	  lit = clause[cli][j];
	  if((lit > 0) == atom[ABS(lit)]){
	    breakcount[ABS(lit)]++;
	    break;
	  }
	}
      }
    }
    
  numocc = numoccurence[MAXATOM+toenforce];
  occptr = occurence[MAXATOM+toenforce];

  occptrind = currentoccurence[MAXATOM+toenforce];


  for(k = 0,i = occptrind[0]; k < numocc ; ++k, i=occptrind[i+1] )
    {
      /* cli = occurence[MAXATOM+toenforce][i]; */
      cli = occptr[i];

      if (++numtruelit[cli] == 1){
	numfalse--;
	false[wherefalse[cli]] =
	  false[numfalse];
	wherefalse[false[numfalse]] =
	  wherefalse[cli];
	/* Increment toflip's breakcount */
	breakcount[toflip]++;

	if (makeflag){
	  /* Decrement the makecount of all vars in the clause */
	  sz = size[cli];
	  litptr = clause[cli];
	  for (j=0; j<sz; j++){
	    lit = clause[cli][j]; 
	    makecount[ABS(lit)]--;
	  }
	}
      }
      else if (numtruelit[cli] == 2){
	/* Find the lit in this clause other than toflip that makes it true,
	   and decrement its breakcount */
	sz = size[cli];
	litptr = clause[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  if( ((lit > 0) == atom[ABS(lit)]) &&
	      (toflip != ABS(lit)) ){
	    breakcount[ABS(lit)]--;
	    break;
	  }
	}
      }
    }
}





int pickrandom(void)
{
  int tofix;

  tofix = false[random()%numfalse];
  return Var(tofix, random()%size[tofix]);
}


int pickbest(void)
{
  int numbreak;
  int tofix;
  int clausesize;
  int i;		
  int best[MAXLENGTH];
  register int numbest;
  register int bestvalue;
  register int var;
 
  tofix = false[random()%numfalse];
  clausesize = size[tofix];
  numbest = 0;
  bestvalue = BIG;

  for (i=0; i< clausesize; i++){
    var = /*ABS(clause[tofix][i])*/ Var(tofix,i) ; 
    numbreak = breakcount[var];
    if (numbreak<=bestvalue){
      if (numbreak<bestvalue) numbest=0;
      bestvalue = numbreak;
      best[numbest++] = var;
    }
  }

  if (bestvalue>0 && (random()%denominator < numerator))
    return /*ABS(clause[tofix][random()%clausesize])*/ Var(tofix,random()%clausesize);

  if (numbest == 1) return best[0];
  return best[random()%numbest];
}

int picknovelty(void)
{
  int var, diff, birthdate;
  int youngest=0, youngest_birthdate, best=0, second_best=0, best_diff, second_best_diff;
  int tofix, clausesize, i, k;

  tofix = false[random()%numfalse];
  clausesize = size[tofix];  

  if (clausesize == 1) return /*ABS(clause[tofix][0]);*/ Var(tofix,0);

  youngest_birthdate = -1;
  best_diff = -BIG;
  second_best_diff = -BIG;

  for(i = 0; i < clausesize; i++){
    var = /*ABS(clause[tofix][i])*/ Var(tofix,i);
    diff = makecount[var] - breakcount[var];
    birthdate = changed[var];
    if (birthdate > youngest_birthdate){
      youngest_birthdate = birthdate;
      youngest = var;
    }
    if (diff > best_diff || (diff == best_diff && changed[var] < changed[best])) {
      /* found new best, demote best to 2nd best */
      second_best = best;
      second_best_diff = best_diff;
      best = var;
      best_diff = diff;
    }
    else if (diff > second_best_diff || (diff == second_best_diff && changed[var] < changed[second_best])){
      /* found new second best */
      second_best = var;
      second_best_diff = diff;
    }
  }

  if (best != youngest){
    return best;
  }
  if ((random()%denominator < numerator)) {
    return second_best;
  }
  return best;
}

int pickrnovelty(void)
{
  int var, diff, birthdate;
  int diffdiff;
  int youngest, youngest_birthdate, best, second_best, best_diff, second_best_diff;
  int tofix, clausesize, i;

  tofix = false[random()%numfalse];
  clausesize = size[tofix];  

  if (clausesize == 1) return /*ABS(clause[tofix][0])*/ Var(tofix,0);
  if ((numflip % 100) == 0) return /*ABS(clause[tofix][random()%clausesize])*/ Var(tofix,random()%clausesize);

  youngest_birthdate = -1;
  best_diff = -BIG;
  second_best_diff = -BIG;

  for(i = 0; i < clausesize; i++){
    var = /*ABS(clause[tofix][i])*/ Var(tofix,i);
    diff = makecount[var] - breakcount[var];
    birthdate = changed[var];
    if (birthdate > youngest_birthdate){
      youngest_birthdate = birthdate;
      youngest = var;
    }
    if (diff > best_diff || (diff == best_diff && changed[var] < changed[best])) {
      /* found new best, demote best to 2nd best */
      second_best = best;
      second_best_diff = best_diff;
      best = var;
      best_diff = diff;
    }
    else if (diff > second_best_diff || (diff == second_best_diff && changed[var] < changed[second_best])){
      /* found new second best */
      second_best = var;
      second_best_diff = diff;
    }
  }
  if (best != youngest) return best;

  diffdiff = best_diff - second_best_diff;

  /*
    if (numerator < 50 && diffdiff > 1) return best;
    if (numerator < 50 && diffdiff == 1){
  */

  if (numerator*2 < denominator && diffdiff > 1) return best;
  if (numerator*2 < denominator && diffdiff == 1){
    if ((random()%denominator) < 2*numerator) return second_best;
    return best;
  }
  if (diffdiff == 1) return second_best;

  /*
    if ((random()%denominator) < 2*(numerator-50)) return second_best;
  */

  if ((random()%denominator) < 2*(numerator-(denominator/2))) return second_best;

  return best;
}


/* hh: the following are variants of novelty and rnovelty that are typically
   more robust than the original heuristics.

   For details, see: 

   Holger H. Hoos and Thomas Stuetzle:
   Local Search Algorithms for SAT: An Empirical Evaluation. 
   Journal of Automated Reasoning, Vol.24, pp.421-481, 2000
   A modified version appeared as a book chapter in: 
   I.P.Gent, H.v.Maaren, T.Walsh, editors, SAT 2000, pp.43-86, IOS Press, 2000. 

   Holger H. Hoos:
   On the Run-time Behaviour of Stochastic Local Search Algorithms for SAT. 
   Proc. of AAAI-99, pp.661-666, MIT Press, 1999

   Holger H. Hoos:
   Stochastic Local Search - Methods, Models, Applications. 
   infix-Verlag, Sankt Augustin, Germany, 1999; ISBN 3-89601-215-0.

   Kautz modified slightly, to fix bug when denominator is not 100

*/

int picknoveltyplus(void)
{
  int var, diff, birthdate;
  int youngest, youngest_birthdate, best, second_best, best_diff, second_best_diff;
  int tofix, clausesize, i;

  tofix = false[random()%numfalse];
  clausesize = size[tofix];  

  if (clausesize == 1) return /*ABS(clause[tofix][0])*/ Var(tofix,0);

  /* hh: inserted modified loop breaker: */
  if ((random()%wp_denominator < wp_numerator)) return /*ABS(clause[tofix][random()%clausesize])*/ Var(tofix,random()%clausesize);

  youngest_birthdate = -1;
  best_diff = -BIG;
  second_best_diff = -BIG;

  for(i = 0; i < clausesize; i++){
    var = ABS(clause[tofix][i]);
    diff = makecount[var] - breakcount[var];
    birthdate = changed[var];
    if (birthdate > youngest_birthdate){
      youngest_birthdate = birthdate;
      youngest = var;
    }
    if (diff > best_diff || (diff == best_diff && changed[var] < changed[best])) {
      /* found new best, demote best to 2nd best */
      second_best = best;
      second_best_diff = best_diff;
      best = var;
      best_diff = diff;
    }
    else if (diff > second_best_diff || (diff == second_best_diff && changed[var] < changed[second_best])){
      /* found new second best */
      second_best = var;
      second_best_diff = diff;
    }
  }
  if (best != youngest) return best;
  if ((random()%denominator < numerator)) return second_best;
  return best;
}

int pickrnoveltyplus(void)
{
  int var, diff, birthdate;
  int diffdiff;
  int youngest=0, youngest_birthdate, best=0, second_best=0, best_diff, second_best_diff;
  int tofix, clausesize, i;


  tofix = false[random()%numfalse];



  clausesize = size[tofix];  

  if (clausesize == 1){
    return ABS(clause[tofix][0]);
  }

/* hh: modified loop breaker: */
  if ((random()%wp_denominator < wp_numerator)){
    return ABS(clause[tofix][random()%clausesize]);
  }

  /*
  if ((numflip % 100) == 0) return ABS(clause[tofix][random()%clausesize]);
  */

  youngest_birthdate = -1;
  best_diff = -BIG;
  second_best_diff = -BIG;

  for(i = 0; i < clausesize; i++){
    var = ABS(clause[tofix][i]);
    diff = makecount[var] - breakcount[var];
    birthdate = changed[var];
    if (birthdate > youngest_birthdate){
      youngest_birthdate = birthdate;
      youngest = var;
    }
    if (diff > best_diff || (diff == best_diff && changed[var] < changed[best])) {
      /* found new best, demote best to 2nd best */
      second_best = best;
      second_best_diff = best_diff;
      best = var;
      best_diff = diff;
    }
    else if (diff > second_best_diff || (diff == second_best_diff && changed[var] < changed[second_best])){
      /* found new second best */
      second_best = var;
      second_best_diff = diff;
    }
  }
  if (best != youngest){
    return best;
  }

  diffdiff = best_diff - second_best_diff;

/*
  if (numerator < 50 && diffdiff > 1) return best;
  if (numerator < 50 && diffdiff == 1){
 */

  if (numerator*2 < denominator && diffdiff > 1){
    return best;
  }
  if (numerator*2 < denominator && diffdiff == 1){
    if ((random()%denominator) < 2*numerator){
      return second_best;
    }
    return best;
  }
  if (diffdiff == 1){
    return second_best;
  }

/*
  if ((random()%denominator) < 2*(numerator-50)) return second_best;
*/

  if ((random()%denominator) < 2*(numerator-(denominator/2))){
    return second_best;
  }
  return best;

}



int picktabu(void)
{
  int numbreak[MAXLENGTH];
  int tofix;
  int clausesize;
  int i;			/* a loop counter */
  int best[MAXLENGTH];	/* best possibility so far */
  int numbest;		/* how many are tied for best */
  int bestvalue;		/* best value so far */
  int noisypick;

  tofix = false[random()%numfalse];
  clausesize = size[tofix];
  for(i = 0;i < clausesize;i++)
    numbreak[i] = breakcount[/*ABS(clause[tofix][i])*/Var(tofix,i)];

  numbest = 0;
  bestvalue = BIG;

  noisypick = (numerator > 0 && random()%denominator < numerator); 
  for (i=0; i < clausesize; i++) {
    if (numbreak[i] == 0) {
      if (bestvalue > 0) {
	bestvalue = 0;
	numbest = 0;
      }
      best[numbest++] = i;
    }
    else if (tabu_length < numflip - changed[/*ABS(clause[tofix][i])*/ Var(tofix,i)]) {
      if (noisypick && bestvalue > 0) { 
	best[numbest++] = i; 
      }
      else {
	if (numbreak[i] < bestvalue) {
	  bestvalue = numbreak[i];
	  numbest = 0;
	}
	if (numbreak[i] == bestvalue) {
	  best[numbest++] = i; 
	}
      }
    }
  }
  if (numbest == 0) return NOVALUE;
  if (numbest == 1) return Var(tofix, best[0]);
  return (Var(tofix, best[random()%numbest]));
}


void initMUSsearch(void){
  int i,j;
  int sizeclause;
  int find;
#ifdef DYNAMIC
  clausedirectory = (int *) malloc(sizeof(int)*(numclause+1));
  backup = (int *) malloc(sizeof(int)*(numclause+1));
  falseorcriticmeter = (int *) malloc(sizeof(int)*(numclause+1));
  clauseinocc = (int **) malloc(sizeof(int *)*(numclause+1));
  defended = (int *) malloc(sizeof(int)*(numclause+1));
  defbackup = (int *) malloc(sizeof(int)*(numclause+1));
  if(sic == TRUE || asmus == TRUE) 
    remclause = (int *) malloc(sizeof(int)*(numclause+1));
  if( asmus == TRUE )
    minfalseasmus = (int *) malloc(sizeof(int)*(numclause+1));
#endif
  for(i = 0;i < numclause;++i)
    {
      clausedirectory[i] = i;
      falseorcriticmeter[i] = 0;
      clauseinocc[i] = (int *) malloc(sizeof(int)*size[i]);
      defended[i] = 0;
    }
  numdefended = 0;
  numclauseinit = numclause;
  if(sic == TRUE || asmus == TRUE) 
    for(i = 0;i < numclause;++i)
      remclause[i] = 0;
}


void initMUSsearch2(void){
  int i,j;
  int numocc;
  int * store;
  numactuallyvar = numatom;
  for(i = 0;i < 2*MAXATOM+1;++i)
    {
       numocc = numoccurence[i];
       store = (int *) malloc(sizeof(int)*(numocc+1));
       poslitinclause[i] = (int *) malloc(sizeof(int)*(numocc));
       for(j = 0;j <= numocc;++j){store[j]= j;}
       currentoccurence[i] = store;
    }
}







void inc_critical_meter(void){
   int i,j;
   int * fausse;
   int sizeclause;
   int poids;
   for(i = 0;i < numfalse ; ++i)
   {
      poids=1;
      for(j=0;j<size[false[i]] && breakcount[ABS(clause[false[i]][j])];++j){poids+=breakcount[ABS(clause[false[i]][j])];}  
      if (j==size[false[i]]){falseorcriticmeter[false[i]]+=poids;}
   }
}


void inc_false_meter(void){
   int i;
   for(i = 0;i < numfalse ; ++i){
      ++falseorcriticmeter[false[i]];
   }
}

int test = 0;
void removeclause(void){
  int i;
  int j;
  int k;
  int cpt;
  int index;
  int lit, pos, following;
  int * removed_clause = (int *) malloc((numclause - numdefended)*sizeof(int));
  int sizerm;
  numvarbackup = numactuallyvar;
  numdefbackup = numdefended;
  numdefended = 0;
  for( i = 0, cpt = 0, withdrawn = 0 ; i < numclause ; ++i )
    {
   
      index = clausedirectory[i];
      backup[i]= clausedirectory[i];
      defbackup[index] = defended[index];
      if(defended[index]){
      	++numdefended;
      }
      if(isCritical(index)){
	/* the clause index is critical  */
	clausedirectory[cpt++] = index ; 

      }else{

	removed_clause[withdrawn++] = index;
        for( j = 0 ; j < size[index] ; ++j){
	  
	  lit = clause[index][j]+MAXATOM;
	  pos = clauseinocc[index][j];
	  if(1+pos<numoccurenceinitial[lit]){
	    following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
	    if(following <numoccurenceinitial[lit]){
	      currentoccurence[lit][pos] = following;
	      clauseinocc[occurence[lit][following]]
		[poslitinclause[lit][following]] =
		pos;
	    }
	  }
	  
	  if(numoccurence[lit]>0){
	    --numoccurence[lit];  
	    if(numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
	      --numactuallyvar;
	    }
	  }
	}
      }
    }
  numbackup = numclause;
  numclause = cpt;
  push(removed_clause,withdrawn);
}

void removeclause_and_parasite(void){
   int i,j,k,lit,nbat;
   int parasiticatoms[numatom];
   removeclause();
   for(i=0, nbat = 0; i < numatom ; ++i)
      if(numoccurence[MAXATOM - i] == 0 && numoccurence[MAXATOM + i] > 0){
         lit = MAXATOM +i;
         for(k = 0,j = currentoccurence[lit][0]; k < numoccurence[lit] && !defended[occurence[lit][j]]; ++k, j = currentoccurence[lit][i+1] );
         if(k>=numoccurence[lit]) parasiticatoms[nbat++] = lit;
      }
      else
         if(numoccurence[MAXATOM + i] == 0 && numoccurence[MAXATOM - i] > 0){
            lit = MAXATOM -i;
            for(k = 0,j = currentoccurence[lit][0]; k < numoccurence[lit] && !defended[occurence[lit][j]]; ++k, j = currentoccurence[lit][i+1] );
            if(k>=numoccurence[lit]) parasiticatoms[nbat++] = lit;
         }
   remove_parasite(parasiticatoms,nbat);
}


void remove_parasite(int * parasiticatoms , int nbat){
   int parasiticclauses[numclause];
   int nbcl;
   int i,j,k,l,m;
   int cli,lit,nb;
   int pos,following,cpt,index;
   for(j=0,nbcl = 0 ; j < nbat ; ++j){
      lit = parasiticatoms[j];
      for(k = 0,i = currentoccurence[lit][0]; k < numoccurence[lit] ; ++k, i = currentoccurence[lit][i+1] ){
         cli = occurence[lit][i];
         if(nbcl > 0){
            for(l=0 ; l < nbcl && cli > parasiticclauses[l] ; ++l) ;
            if(l < nbcl && cli < parasiticclauses[l] ){
               for(m = nbcl ; m > l ; --m){
                  parasiticclauses[m] = parasiticclauses[m-1];
               }
               parasiticclauses[l] = cli;
               ++nbcl;
            }
         }else{
            parasiticclauses[nbcl++] = cli;
         }
      }
   }
   if(nbcl>0){
      for(i = 0,cpt = 0 ,m = 0; i < numclause ; ++i)
      {
         index = clausedirectory[i];
         if(index != parasiticclauses[m]){
            clausedirectory[cpt++] = index;
            
         }else{
            ++m;
            for( j = 0 ; j < size[index] ; ++j){
               lit = clause[index][j]+MAXATOM;
               pos = clauseinocc[index][j];
               if(1+pos<numoccurenceinitial[lit]){
                  following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
                  if(following <numoccurenceinitial[lit]){
                     currentoccurence[lit][pos] = following;
                     clauseinocc[occurence[lit][following]]
                        [poslitinclause[lit][following]] =
                        pos;
                  }
               }
               if(numoccurence[lit]>0){
                  --numoccurence[lit]; 
                  if( numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
                     --numactuallyvar;
                  }
               } 
            }
         }
      }
      numclause = cpt ;
   }
}

void resetclause(void){
  thresholdcalculator();
  rmclause();
}


void maxminmeter1(int * max,int * min){
  int i=0;
  *max = clausedirectory[i];
  *min = clausedirectory[i];
  for(; i < numclause ; ++i)
    {
	if(falseorcriticmeter[*max] < falseorcriticmeter[clausedirectory[i]])
	  *max = clausedirectory[i];
	else if(falseorcriticmeter[*min] > falseorcriticmeter[clausedirectory[i]])
	  *min = clausedirectory[i];
    }
}




void oldthresholdcalculator(void){
  int max,min,nbClauseSelectionnee,i,j,k;
  int minScore,maxscore;
  maxminmeter1(&max,&min);
  minScore=falseorcriticmeter[min];
  maxscore=falseorcriticmeter[max];
  nbClauseSelectionnee=numclause;
  j=0;
  while(nbClauseSelectionnee==numclause){
     nbClauseSelectionnee=0;
     switch(j++){
        case 0 :
        {
           if ((optimization) && (!(falseClause))){
              criticalpoint=minScore;
           }else criticalpoint=minScore+((cutoff/numclause)/iteration); 
           break;
        }
        case 1: {criticalpoint=(1023*criticalpoint+((minScore+maxscore)>>1))>>10; break;}
        case 2 : {criticalpoint=(511*criticalpoint+((minScore+maxscore)>>1))>>9; break;}
        case 3 : {criticalpoint=(255*criticalpoint+((minScore+maxscore)>>1))>>8; break;}
        case 4 : {criticalpoint=(127*criticalpoint+((minScore+maxscore)>>1))>>7; break;}
        case 5 : {criticalpoint=(63*criticalpoint+((minScore+maxscore)>>1))>>6; break;}
        case 6 : {criticalpoint=(31*criticalpoint+((minScore+maxscore)>>1))>>5; break;}
        case 7 : {criticalpoint=(15*criticalpoint+((minScore+maxscore)>>1))>>4; break;}
        case 8 : {criticalpoint=(7*criticalpoint+((minScore+maxscore)>>1))>>3; break;}
        case 9 : {criticalpoint=(3*criticalpoint+((minScore+maxscore)>1))>>2; break;}
        case 10 : {criticalpoint=(criticalpoint+((minScore+maxscore)>>1))>>1; break;}
        case 11 : {criticalpoint=(minScore+maxscore)>>1; break;}
        case 12 : {criticalpoint=maxscore+1; break;}
           /* If all the clauses are defended, we put nbClauseSelectionnee and criticalpoint to -1, in order to exit that loop  */
        default :{if (numclause==numdefended) {nbClauseSelectionnee=-1; criticalpoint=-1;break;} else {fprintf(stderr,"Unknown error occurs. variable j in genereFichierMin is not correct (%i instead of being between 0 and 12)\n",j); exit(-1);}}
     }
     for(i=0,k=clausedirectory[i];i<numclause;i++,k=clausedirectory[i]) if (isCritical(k)) nbClauseSelectionnee++; 
  }
  ++testnumber;
}


void maxminmeter2(int * max,int * min){
   int i=0;
   for(i=0;i<numclause && defended[clausedirectory[i]];++i);
   *max = clausedirectory[i];
   *min = clausedirectory[i];
   for(; i < numclause ; ++i)
   {
      if(!defended[clausedirectory[i]]){
         if(falseorcriticmeter[*max] < falseorcriticmeter[clausedirectory[i]])
            *max = clausedirectory[i];
         else if(falseorcriticmeter[*min] > falseorcriticmeter[clausedirectory[i]])
            *min = clausedirectory[i];
      }
    }
}



void newthresholdcalculator(void){
  int max,min,nbClauseSelectionnee,i,j,k;
  int minScore,maxscore;
  maxminmeter2(&max,&min);
  minScore=falseorcriticmeter[min];
  maxscore=falseorcriticmeter[max];
  criticalpoint = falseorcriticmeter[min] + ((newth*cutoff) / ((++testnumber+iteration)*numclause)) ;
}


int isCritical(int clause){
   return (defended[clause] || falseorcriticmeter[clause] > criticalpoint || withdrawn >= raising);
}


char * convert_integer_into_string(int n){
   int tmp;
   char r;
   int cpt;
   char * c;
   int pos = 1;
   if(n<0){
      pos = 0;
      n*=-1;
   }
   for(cpt = (pos)? 2 : 3 ,tmp=n;tmp/=10;++cpt);
   c=(char *)malloc(cpt*sizeof(char));
   if(!c){
      exit(12);
   }
   c[--cpt]='\0';
   for(tmp=n;tmp;tmp/=10){
      r=tmp%10;
      c[--cpt]=r+'0';
   }
   r=tmp%10;
   c[--cpt]=r+'0';
   if(!pos)c[0] = '-';
   return c;
}



void directory_into_file(char * file,const int * directory,const int taille){
   int file1;
   int i,j,index;
   char * buf;
   file1 = open(file,O_WRONLY|O_CREAT|O_EXCL,00600);
   if(file1 == -1 && errno == EEXIST){
      if(remove(file)){
         perror("failed while erasing the file");
         exit(-2);
      }
      file1 = open(file,O_WRONLY|O_CREAT,00600); 
   }
   if(file1 == -1){
      perror("failed while creating the file");
      exit(-1);
   }
   write(file1,"p cnf ",strlen("p cnf "));
   buf = convert_integer_into_string(numatom);  
   write(file1,buf,strlen(buf));
   write(file1," ",1);
   buf = convert_integer_into_string(taille);
   write(file1,buf,strlen(buf));
   write(file1,"\n",1);
   for( i = 0 ; i < taille ; ++i)
   {
      index = directory[i];
      
      for( j = 0 ; j < size[index] ; ++j)
      {
         buf = convert_integer_into_string(clause[index][j]);
         write(file1,buf,strlen(buf));
         write(file1," ",1);
      }
      write(file1,"0 \n",3);
    }
   close(file1);
}




void DL_verification(void){
  int ret;
  char * file;
  if(testnumber == 1 && nowithdrawnclause < numwithdrawn){
    testnumber = 0;
    ++iteration;
    nowithdrawnclause++;
  }else{
    ret = unsat(backupfile);
    if(ret){
      nofound = 0;
      if(print == TRUE){
         printf("------------------------------------------------------------------\n");
         printf("%d iterations\n",testnumber);
         printf("\nExtracted Sub-formula :\n");
         printf("#Var %d\n#Clause %d\n\n",numvarbackup,numbackup);
      }
      
      if(finetune == TRUE || sic == TRUE || asmus == TRUE){
	
	fineTune_extraction();
	allot_name_file(&file);
	directory_into_file(file,backup,numbackup);
	if(print == TRUE){
	  printf("Extracted MUS :\n");
	  printf("#Var %d\n#Clause %d\n\n",numvarbackup,numbackup);
	}
      }
    }else{
	correction_clause(); 
	nofound = 0;
    }
  }
}


void fineTune_verification(int * directory,int * num,int index){


  if(!unsat(backupfile)){
    directory[(*num)++] = index;
  }

}




void fineTune_extraction(void){
  
  int i,j;
  int index;
  int directory[numbackup];
  int num = 0; 
 
  int tmpdir[numbackup];
  int tmpnum;


  for( i = 0 ; i < numbackup ;++i){

    index = backup[i];


    if(defbackup[index]){
      directory[num++] = index;
    
    }else{
      for( j = 0 , tmpnum = 0; j < num ; ++j)tmpdir[tmpnum++] = directory[j];
      for( j = i+1 ; j < numbackup ; ++j)tmpdir[tmpnum++] = backup[j];

      directory_into_file(backupfile,tmpdir,tmpnum);

      fineTune_verification(directory,&num,index);

      
    }
    
   
  }




  for( i = 0 ; i < num ; ++i)
    {
      
      backup[i] = directory[i];
   
    }


  numbackup = num;

  
}



void print_try(void)
{

  if(print == TRUE){
     //printf("\n--------------------------------------\n");
     //printf("# try %d                             #\n",iteration);
     printf("------------------------------------------------------------------\n");
  }



}



void allot_name_file(char ** file)
{

  char * num;
  int taille;

  if ( sic == FALSE )
    *file = strdup(backupfile);
  else
    {
      num = convert_integer_into_string(++siccpt);
      taille = strlen(num) + strlen(backupfile) + 5;
      *file = (char *) malloc(sizeof(char)*taille);
      strcpy(*file,backupfile);
      strcat(*file,"_mus");
      strcat(*file,num);
      printf("MUS#%d written in file \"%s\"\n",siccpt,*file);
    }

}




void reset_for_SIC_search(void)
{

  int i;
  int j;
  int cpt;
  int k;
  int lit, pos, following;
  char * file;
  int taille;

    
  if(nofound)
    {
	
      for( i = 0, cpt = 0 ; i < numclause ; ++i )
	{
	     
	  if(remclause[i] == 0)
	    {
	      clausedirectory[cpt++] = i;
	    }
	  else
	    {
	
	      for( j = 0 ; j < size[i] ; ++j){
		    
		lit = clause[i][j]+MAXATOM;
		pos = clauseinocc[i][j];
		if(1+pos<numoccurenceinitial[lit]){
		  following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
		  if(following < numoccurenceinitial[lit]){
		    currentoccurence[lit][pos] = following;
		    clauseinocc[occurence[lit][following]]
		      [poslitinclause[lit][following]] =
		      pos;
		    
		  }
		}
		    
		if(numoccurence[lit]>0){
		  --numoccurence[lit];  
		  if(numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
		    --numactuallyvar;
		  }
		}
		    
	      }
	    }
	}
      numclause = cpt;
	
    }
  else
    {
      for( i = 0, cpt = 0 , k = 0 ; i < numclause ; ++i )
	{
	    
	  if(i != backup[k] && remclause[i] == 0){
	    clausedirectory[cpt++] = i;
	  }
	  else
	    {
		
	      if(remclause[i] == 0){
		++k;
		remclause[i] = 1;
	      }
	      for( j = 0 ; j < size[i] ; ++j){
		
		lit = clause[i][j]+MAXATOM;
		pos = clauseinocc[i][j];
		
		if(1+pos<numoccurenceinitial[lit]){
		  following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
		
		  if(following <numoccurenceinitial[lit]){
		    currentoccurence[lit][pos] = following;
		    clauseinocc[occurence[lit][following]]
		      [poslitinclause[lit][following]] =
		      pos;
		
		  }
		    
		}
	
		if(numoccurence[lit]>0){
		  --numoccurence[lit];  
		  if(numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
		    --numactuallyvar;
		  }
		}
		
	      }

	    }
	}
      numclause = cpt;

	      
      taille = strlen(backupfile) + 5;
      file = (char *)malloc(sizeof(char)*taille);
      strcpy(file,backupfile);
      strcat(file,"_sat");
      
      if(cpt>0)
	{  
	  directory_into_file(file,clausedirectory,numclause);
	    
	  if(unsat(file))
	    {
	      nofound = 1;
	      iteration = 1;
	      testnumber = 0;
	      printf("\nNEW SIC ITERATION :\n");
	      printf("-------------------\n\n");
	    }
	  else{
	    printf("\nResult :\n");
	    printf("@@@@@@@@\n");
	    printf("\nSIC written in file \"%s\"\n",file);
	    remove(backupfile);
	      
	  }
	}
      else{
	printf("\nResult :\n");
	printf("@@@@@@@@\n");
	printf("\nSIC is empty\n");
	remove(file);
	remove(backupfile);
	  
      }

	
    }

    
}


void reset_for_ASMUS_search(void)
{
  int i;
  int j;
  int k;
  int cpt;
  int lit, pos, following;
  char * file;
  int taille;
  int index;

 
  for( i = 0, cpt = 0 ; i < numclause ; ++i )
    {
     
      if(remclause[i] == 0)
	{
	  clausedirectory[cpt++] = i;
	}
      else
	{
	  for( j = 0 ; j < size[i] ; ++j){
	    
	    lit = clause[i][j]+MAXATOM;
	    pos = clauseinocc[i][j];
	    if(1+pos<numoccurenceinitial[lit]){
	      following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
	      if(following < numoccurenceinitial[lit]){
		currentoccurence[lit][pos] = following;
		clauseinocc[occurence[lit][following]]
		  [poslitinclause[lit][following]] =
		  pos;
		
	      }
	    }
	    
	    if(numoccurence[lit]>0){
	      --numoccurence[lit];  
	      if(numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
		--numactuallyvar;
	      }
	    }
	    
	  }
	}
    }
  numclause = cpt;

  
  if(!nofound)
    {  
      index = clause_of_the_lowest_layer();
      
      printf("\nASMUS search: withdrawn clause #%d\n\n",index); 
  

      remclause[index] = 1;
  
      for( j = 0 ; j < size[index] ; ++j){
		
	lit = clause[index][j]+MAXATOM;
	pos = clauseinocc[index][j];
	
	if(1+pos<numoccurenceinitial[lit]){
	  following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
		
	  if(following <numoccurenceinitial[lit]){
	    currentoccurence[lit][pos] = following;
	    clauseinocc[occurence[lit][following]]
	      [poslitinclause[lit][following]] =
	      pos;
	    
	  }
	  
	}
	
	if(numoccurence[lit]>0){
	  --numoccurence[lit];  
	  if(numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
	    --numactuallyvar;
	  }
	}
	
      }

      for(i=0 ; clausedirectory[i] != index && i < numclause; ++i);
      for(; i < numclause;++i)
	clausedirectory[i] = clausedirectory[i+1];


      --numclause;

       
      taille = strlen(backupfile) + 5;
      file = (char *)malloc(sizeof(char)*taille);
      strcpy(file,backupfile);
      strcat(file,"_sat");
      
      if(numclause > 0)
	{  
	  directory_into_file(file,clausedirectory,numclause);
	  
	  if(unsat(file))
	    {
	      nofound = 1;
	      iteration = 1;
	      testnumber = 0;
	      printf("\nNEW ASMUS ITERATION :\n");
	      printf("-------------------\n\n");

	      for(k = 0, i = clausedirectory[k];k < numclause; i = clausedirectory[++k] )
		minfalseasmus[i] = numclause;

	    }
	  else{
	    printf("\nResult :\n");
	    printf("@@@@@@@@\n");
	    printf("\nASMUS written in file \"%s\"\n",file);
	    remove(backupfile);
	    
	  }
	}
      else{
	printf("\nResult :\n");
	printf("@@@@@@@@\n");
	printf("\nASMUS is empty\n");
	remove(file);
	remove(backupfile);
	
      }
    }


}




int clause_of_the_lowest_layer(void)
{
  int i,k;
  int max = backup[0];
  
  for(k = 1, i = backup[k];k < numbackup; i = backup[++k] )
    {
      if((minfalseasmus[i] > minfalseasmus[max]) || (minfalseasmus[i]== minfalseasmus[max] && falseorcriticmeter[i] < falseorcriticmeter[max]))
	max = i;
    }


  return max;

}


int unsat(char * file)
{
  int pid;
  int ret = 1;
  int status;
  

  pid = fork ();

  if (pid > 0) {
    waitpid(pid,&status,0);

    if(! WIFEXITED(status)){
      perror("file control failed");
      exit(-3);
    }
    
    ret = WEXITSTATUS(status);

  }else{
    if(pid == 0) {
      int d = open("/dev/null",O_WRONLY,00700);
      close(1);
      if(dup(d)==-1){
	perror("duplication failed\n");
	exit(-4);
      }
      close(d);
      
      execl("zchaff","zchaff",file,NULL);
      perror("file control failed");
      exit(-2);     


    }else{
      perror("file control failed");
      exit(-1);
    }
  }

  return !ret;


}


void  print_Var_Clause(){

  if(firstloop){
    printf("#Var %d\n",numatom);
    printf("#Clause %d\n\n\n",numclause);
    firstloop = 0;
  }
}


void  init_random_seed(){


  gettimeofday(&tv,&tzp);
  seed = (unsigned int)((( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec);




}



void init_numtry_and_numrun()
{
  
  numtry = 0;
  if(asmus == TRUE && testnumber == 0)
    numrun = numactuallyvar;
  else
    numrun = 10 + ((numclause>>10)*sqrt(iteration));   
}




void flipatom_and_count_defended()
{
  flipatom((pickcode[heuristic])());
  if(numfalse ==1){
    if(!defended[false[0]]) ++numdefended;
    defended[false[0]] = 1;
  }else
    incfalseorcriticmeter();

}


void flipatom_and_count_defended_and_count_minfalseasmus(void)
{
  int i;

  flipatom_and_count_defended();
  for(i=0 ; i < numfalse ;++i)
    if(minfalseasmus[false[i]] > numfalse)
      minfalseasmus[false[i]] = numfalse;
  
}


void control_of_end_of_try_loop()
{
  int ret;
  int * tmp1,tmp2;
    
  if(numdefended == numclause && numclause > 0){
    if(numfalse <= target){
      finished = 1;
      directory_into_file(backupfile,backup,numbackup);	
    }else{
      directory_into_file(backupfile,clausedirectory,numclause);	
      ret = unsat(backupfile);
      
      if(!ret){
	finished = 1;
	directory_into_file(backupfile,backup,numbackup);
      }else{
	finished = 2;
	numvarbackup = numactuallyvar;
      }
    }
      
   
  }else{
    if(numfalse <= target){
      finished = 1;
      directory_into_file(backupfile,backup,numbackup);
      if(optimization == FALSE && testnumber > 1){
	finished = 3;
	seed += OPTI;
	optimization = TRUE;
	pop(&tmp1,&tmp2);
      }
    }
   
  }
}




void print_stat_of_try()
{
  if(print == TRUE)
    printf("%d\t\t(try %d\t\tflip %lld\t#protected %d)\n",numclause,numrun,cutoff,numdefended); 
          
}


void print_result()
{
  if(sic != TRUE && asmus != TRUE){
     //printf("\nResults :\n");
     //printf("@@@@@@@@\n\n");
     //printf("Extracted MUS :\n");
     //printf("#Var %d\n",numvarbackup);
     //printf("#Clause %d\n",numbackup);
     //printf("------------------------------------------------------------------\n\n");
     ;
  }
}





void reset_sic(void)
{
  resetprob();
  reset_for_SIC_search();
}


void reset_asmus(void)
{
 resetprob();
 reset_for_ASMUS_search();
 
}

void nullmethod(){
  int i,k;
  for(i = 0, k = clausedirectory[i] ;i < numclause;k = clausedirectory[++i]){
    falseorcriticmeter[k] =0;
  }  
}


void agingmethod(){
  int i,k;
  for(i = 0, k = clausedirectory[i] ;i < numclause;k = clausedirectory[++i]){
    falseorcriticmeter[k] = (int) (falseorcriticmeter[k] * testnumber) / cutoff ;
  }  
}


void back_return(){

  int i,j,k,cpt,lit,pos,following;

  resetprob();

  for( i = 0,k = 0, cpt = 0 ; i < numclause ; ++i ){
    defended[i] = defbackup[i];
    if(i == backup[k]){
      clausedirectory[cpt++] = i;
      ++k;
    }else{
      for( j = 0 ; j < size[i] ; ++j){
	    
	lit = clause[i][j]+MAXATOM;
	pos = clauseinocc[i][j];
	if(1+pos<numoccurenceinitial[lit]){
	  following = currentoccurence[lit][1+currentoccurence[lit][pos]]; 
	  if(following < numoccurenceinitial[lit]){
	    currentoccurence[lit][pos] = following;
	    clauseinocc[occurence[lit][following]]
	      [poslitinclause[lit][following]] =
	      pos;
	  }
	}
	    
	if(numoccurence[lit]>0){
	  --numoccurence[lit];  
	  if(numoccurence[lit] == 0 && numoccurence[2*MAXATOM-lit] == 0){
	    --numactuallyvar;
	  }
	}
	
      }
    }
  }
  numclause = cpt;
  numdefended = numdefbackup;

 

}

void level_init(){

  switch(level){
  case 1 :
    resetmeter = agingmethod;
    rmclause = removeclause_and_parasite;
    break;
  case 2 :
    rmclause = removeclause_and_parasite;
    break;
  case 3 :
    thresholdcalculator = newthresholdcalculator;
    newth = 5;
    rmclause = removeclause_and_parasite;
    break;
  case 4 : 
    break;
  case 5 : 
    thresholdcalculator = newthresholdcalculator;
    newth = 5;
    rmclause = removeclause_and_parasite;
    resetmeter = agingmethod; 
    break;
  case 6 :
    thresholdcalculator = newthresholdcalculator;
    newth = 1;
    rmclause = removeclause_and_parasite;
    break;
  case 7 :
    thresholdcalculator = newthresholdcalculator;
    newth = 1;
    resetmeter = agingmethod; 
    rmclause = removeclause_and_parasite;
    break; 
  default :
    fprintf(stderr,"bad level\n");
    exit(-1);
    break;
  }



}


void correction_clause(){
  
  int i;
  int ret;
  int * removed_clause;
  int sz;
  char * file;
 

  if(numbackup != numclause){
    pop(&removed_clause,&sz);
  }
  do{
    
    pop(&removed_clause,&sz);
    for(i = 0;i<sz;++i){
      backup[numbackup++] = removed_clause[i];
    }
    directory_into_file(backupfile,backup,numbackup);
    ret = unsat(backupfile);
  }while(!ret && numbackup != numclauseinit);
  if(print == TRUE){
     printf("------------------------------------------------------------------\n");
     printf("%d iterations\n",testnumber);
     printf("\nExtracted Sub-formula :\n");
     printf("#Var %d\n#Clause %d\n\n",numvarbackup,numbackup);
  }
  if(finetune == TRUE || sic == TRUE || asmus == TRUE){
     fineTune_extraction();
     allot_name_file(&file);
     directory_into_file(file,backup,numbackup);
     if(print == TRUE){
        printf("Extracted MUS :\n");
        printf("#Var %d\n#Clause %d\n\n",numvarbackup,numbackup);
     }
  }
  quicksort(numbackup,backup);
}


void quicksort(int sz,int * directory){

  int * dirtmp = (int *) malloc(sz * sizeof(int));
  int pivot = directory[((int)sz/2)];
  int i,inf,sup;
  
  if(sz>1){
    for(i=0,inf=0,sup=sz-1;i<sz;++i)
      if(directory[i] < pivot) dirtmp[inf++] = directory[i]; else if(directory[i] > pivot) dirtmp[sup--] = directory[i];
   
    dirtmp[sup] = pivot;
    
    quicksort(inf,dirtmp);
    quicksort(sz-sup-1 ,dirtmp+sup+1);
    
    for(i=0;i<sz;++i){
      directory[i] = dirtmp[i];
    }
    
    free(dirtmp);
  }
}


void init_stack(){

  top = 0;
  sizespace = 0;

}


void push(int * cl, int sz){

  int i;
  int ** sttmp, * sztmp;

  if(top >= sizespace){
    sizespace += BLOC;
    sttmp = (int **)malloc(sizespace*sizeof(int *));
    sztmp = (int *) malloc(sizespace*sizeof(int));
    
    if(!sttmp && !sztmp) {
      fprintf(stderr,"memory problem\n");
      exit(-1);
    } 

    for(i=0 ; i < top;++i){
      sttmp[i] = stack[i];
      sztmp[i] = sizestack[i];
    }

    if(top>0){
      free(stack);
      free(sizestack);
    }

    stack = sttmp;
    sizestack = sztmp;
   
  }

  stack[top] = cl;
  sizestack[top++] = sz;


}

void pop(int ** cl, int * sz){

  *cl = stack[--top];
  *sz = sizestack[top];

}
